import Foundation
import UIKit
import CoreData
import DZNEmptyDataSet
import ReadingList_Foundation

enum ListBooksSource {
    case controller(NSFetchedResultsController<Book>)
    case orderedSet(NSOrderedSet)

    func numberOfBooks() -> Int {
        switch self {
        case .controller(let controller):
            return controller.sections![0].numberOfObjects
        case .orderedSet(let set):
            return set.count
        }
    }

    func book(at indexPath: IndexPath) -> Book {
        switch self {
        case .controller(let controller):
            return controller.object(at: indexPath)
        case .orderedSet(let set):
            return set.object(at: indexPath.row) as! Book
        }
    }
}

class ListBookTable: UITableViewController {

    var list: List!
    private var cachedListNames: [String]!
    private var ignoreNotifications = false
    private var searchController: UISearchController!
    private var listBookSource: ListBooksSource!

    var showSearchBarOnAppearance: Bool = false {
        didSet { navigationItem.hidesSearchBarWhenScrolling = !showSearchBarOnAppearance }
    }

    private var listNameField: UITextField? {
        get { return navigationItem.titleView as? UITextField }
        set { navigationItem.titleView = newValue }
    }

    private var listNameFieldDefaultText: String {
        return "\(list.name)⌄"
    }

    private var defaultPredicate: NSPredicate {
        return NSPredicate(format: "%@ IN %K", list, #keyPath(Book.lists))
    }

    override func viewDidLoad() {
        super.viewDidLoad()
        tableView.register(BookTableViewCell.self)
        tableView.register(BookTableHeader.self)

        cachedListNames = List.names(fromContext: PersistentStoreManager.container.viewContext)
        navigationItem.title = list.name
        navigationItem.rightBarButtonItem = editButtonItem

        tableView.emptyDataSetSource = self
        tableView.emptyDataSetDelegate = self

        buildBookSource()

        searchController = UISearchController(filterPlaceholderText: "Filter List")
        searchController.searchResultsUpdater = self
        navigationItem.searchController = searchController

        NotificationCenter.default.addObserver(self, selector: #selector(objectContextChanged(_:)),
                                               name: .NSManagedObjectContextObjectsDidChange,
                                               object: PersistentStoreManager.container.viewContext)
        monitorThemeSetting()
    }

    private func buildBookSource() {
        // We cannot use a fetched results controller when ordering by the underlying ordered predicate order.
        // Instead, we just use the ordered set as our source.
        if list.order == .listCustom {
            listBookSource = .orderedSet(list.books)
        } else {
            listBookSource = .controller(buildResultsController())
        }
    }

    private func buildResultsController() -> NSFetchedResultsController<Book> {
        let fetchRequest = NSManagedObject.fetchRequest(Book.self, batch: 50)
        fetchRequest.predicate = defaultPredicate
        fetchRequest.sortDescriptors = list.order.sortDescriptors
        let controller = NSFetchedResultsController(fetchRequest: fetchRequest,
                                                    managedObjectContext: PersistentStoreManager.container.viewContext,
                                                    sectionNameKeyPath: nil, cacheName: nil)
        controller.delegate = self
        try! controller.performFetch()
        return controller
    }

    override func viewDidAppear(_ animated: Bool) {
        super.viewDidAppear(animated)

        // Ensure that hidesSearchBarWhenScrolling is always true when the view appears.
        // Works in conjunction with showSearchBarOnAppearance.
        navigationItem.hidesSearchBarWhenScrolling = true
    }

    override func initialise(withTheme theme: Theme) {
        super.initialise(withTheme: theme)
        if let listNameField = listNameField {
            listNameField.textColor = theme.titleTextColor
        }
    }

    private func listTextField() -> UITextField {
        guard let navigationBar = navigationController?.navigationBar else { preconditionFailure() }
        let theme = UserDefaults.standard[.theme]
        let textField = UITextField(frame: navigationBar.frame.inset(by: UIEdgeInsets(top: 0, left: 115, bottom: 0, right: 115)))
        textField.text = listNameFieldDefaultText
        textField.textAlignment = .center
        textField.font = UIFont.systemFont(ofSize: 17.0, weight: .semibold)
        textField.textColor = theme.titleTextColor
        textField.keyboardAppearance = theme.keyboardAppearance
        textField.enablesReturnKeyAutomatically = true
        textField.returnKeyType = .done
        textField.delegate = self
        textField.addTarget(self, action: #selector(self.configureNavigationItem), for: .editingChanged)
        return textField
    }

    private func canUpdateListName(to name: String) -> Bool {
        guard !name.isEmptyOrWhitespace else { return false }
        return name == list.name || !cachedListNames.contains(name)
    }

    @discardableResult private func tryUpdateListName(to name: String) -> Bool {
        if canUpdateListName(to: name) {
            UserEngagement.logEvent(.renameList)
            list.name = name
            list.managedObjectContext!.saveAndLogIfErrored()
            return true
        } else {
            return false
        }
    }

    override func setEditing(_ editing: Bool, animated: Bool) {
        super.setEditing(editing, animated: animated)

        // If we go from editing to not editing, and we are (/were) editing the title text field, then
        // save the update (if we can), and stop editing it.
        if !editing, let listNameField = listNameField, listNameField.isEditing {
            if let proposedName = listNameField.text, list.name != proposedName {
                tryUpdateListName(to: proposedName)
            }
            listNameField.endEditing(true)
        }
        configureNavigationItem()
        reloadHeaders()
    }

    override func tableView(_ tableView: UITableView, viewForHeaderInSection section: Int) -> UIView? {
        guard section == 0 && tableView.numberOfRows(inSection: 0) > 0 else { return nil }
        let header = tableView.dequeue(BookTableHeader.self)
        header.presenter = self
        header.onSortChanged = sortOrderChanged
        configureHeader(header, at: section)
        return header
    }

    @objc private func configureNavigationItem() {
        guard let editDoneButton = navigationItem.rightBarButtonItem else { assertionFailure(); return }
        editDoneButton.isEnabled = {
            if let listNameField = listNameField {
                if !listNameField.isEditing { return true }
                if let newName = listNameField.text, canUpdateListName(to: newName) { return true }
                return false
            }
            return true
        }()
        searchController.searchBar.isEnabled = !isEditing
        if isEditing {
            if listNameField == nil {
                listNameField = listTextField()
                navigationItem.title = nil
            }
        } else {
            listNameField = nil
            navigationItem.title = list.name
        }
    }

    private func sortOrderChanged() {
        if searchController.isActive {
            // We don't allow sort order change while the search controller is active; if it does, stop the search.
            assertionFailure()
            searchController.isActive = false
        }
        buildBookSource()
        tableView.reloadData()

        // Put the top row at the "middle", so that the top row is not right up at the top of the table
        tableView.scrollToRow(at: IndexPath(row: 0, section: 0), at: .middle, animated: false)
        UserEngagement.logEvent(.changeListSortOrder)
    }

    @objc private func objectContextChanged(_ notification: Notification) {
        guard !ignoreNotifications else { return }
        guard let userInfo = notification.userInfo else { return }

        if (userInfo[NSDeletedObjectsKey] as? NSSet)?.contains(list) == true {
            // If the list was deleted, pop back. This can't happen through any normal means at the moment.
            navigationController?.popViewController(animated: false)
            return
        }

        // Repopulate the list names cache
        cachedListNames = List.names(fromContext: PersistentStoreManager.container.viewContext)

        // If we are not using a controller, reload the table
        if case .orderedSet = listBookSource! {
            tableView.reloadData()
        }
    }

    private func ignoringSaveNotifications(_ block: () -> Void) {
        ignoreNotifications = true
        block()
        ignoreNotifications = false
    }

    override func numberOfSections(in tableView: UITableView) -> Int {
        switch listBookSource! {
        case .controller(let controller):
            return controller.sections!.count
        case .orderedSet(let set):
            return set.isEmpty ? 0 : 1
        }
    }

    override func tableView(_ tableView: UITableView, numberOfRowsInSection section: Int) -> Int {
        guard section == 0 else { assertionFailure(); return 0 }
        return listBookSource.numberOfBooks()
    }

    override func tableView(_ tableView: UITableView, cellForRowAt indexPath: IndexPath) -> UITableViewCell {
        let cell = tableView.dequeue(BookTableViewCell.self, for: indexPath)
        let book = listBookSource.book(at: indexPath)
        cell.initialise(withTheme: UserDefaults.standard[.theme])
        cell.configureFrom(book, includeReadDates: false)
        return cell
    }

    override func tableView(_ tableView: UITableView, didSelectRowAt indexPath: IndexPath) {
        performSegue(withIdentifier: "showDetail", sender: indexPath)
    }

    override func shouldPerformSegue(withIdentifier identifier: String, sender: Any?) -> Bool {
        return !tableView.isEditing
    }

    override func tableView(_ tableView: UITableView, editActionsForRowAt indexPath: IndexPath) -> [UITableViewRowAction]? {
        return [UITableViewRowAction(style: .destructive, title: "Remove") { _, indexPath in
            let bookToRemove = self.listBookSource.book(at: indexPath)
            self.list.removeBooks(NSSet(object: bookToRemove))
            // Ignore save notifications, so we don't reload the table when using the set: we will remove the row manually
            self.ignoringSaveNotifications {
                self.list.managedObjectContext!.saveAndLogIfErrored()
            }
            if case .orderedSet(let set) = self.listBookSource! {
                let mutableSet = set.mutableCopy() as! NSMutableOrderedSet
                mutableSet.remove(bookToRemove)
                self.listBookSource = .orderedSet(mutableSet)
            }
            if case .orderedSet(let set) = self.listBookSource! {
                if set.isEmpty {
                    tableView.deleteSections(IndexSet(arrayLiteral: 0), with: .automatic)
                    tableView.reloadEmptyDataSet()
                } else {
                    tableView.deleteRows(at: [indexPath], with: .automatic)
                }
            }
            UserEngagement.logEvent(.removeBookFromList)
        }]
    }

    override func tableView(_ tableView: UITableView, canMoveRowAt indexPath: IndexPath) -> Bool {
        guard !searchController.isActive else { return false }
        return list.order == .listCustom && list.books.count > 1
    }

    override func tableView(_ tableView: UITableView, moveRowAt sourceIndexPath: IndexPath, to destinationIndexPath: IndexPath) {
        guard list.order == .listCustom else { assertionFailure(); return }
        guard sourceIndexPath != destinationIndexPath else { return }
        ignoringSaveNotifications {
            var books = list.books.map { $0 as! Book }
            let movedBook = books.remove(at: sourceIndexPath.row)
            books.insert(movedBook, at: destinationIndexPath.row)
            list.books = NSOrderedSet(array: books)
            list.managedObjectContext!.saveAndLogIfErrored()
        }
        UserEngagement.logEvent(.reorederList)
    }

    override func prepare(for segue: UIStoryboardSegue, sender: Any?) {
        if let detailsViewController = (segue.destination as? UINavigationController)?.topViewController as? BookDetails {
            guard let senderIndex = sender as? IndexPath else { preconditionFailure() }
            let book = listBookSource.book(at: senderIndex)
            detailsViewController.book = book
        }
    }
}

extension ListBookTable: NSFetchedResultsControllerDelegate {
    func controllerWillChangeContent(_: NSFetchedResultsController<NSFetchRequestResult>) {
        tableView.beginUpdates()
    }

    func controllerDidChangeContent(_: NSFetchedResultsController<NSFetchRequestResult>) {
        tableView.endUpdates()
        reloadHeaders()
    }

    func controller(_ controller: NSFetchedResultsController<NSFetchRequestResult>, didChange anObject: Any, at indexPath: IndexPath?, for type: NSFetchedResultsChangeType, newIndexPath: IndexPath?) {
        tableView.controller(controller, didChange: anObject, at: indexPath, for: type, newIndexPath: newIndexPath)
    }

    func controller(_ controller: NSFetchedResultsController<NSFetchRequestResult>, didChange sectionInfo: NSFetchedResultsSectionInfo, atSectionIndex sectionIndex: Int, for type: NSFetchedResultsChangeType) {
        tableView.controller(controller, didChange: sectionInfo, atSectionIndex: sectionIndex, for: type)
    }
}

extension ListBookTable: UISearchResultsUpdating {
    private func getSearchPredicate() -> NSPredicate? {
        guard let searchTerms = searchController.searchBar.text else { return nil }
        if searchTerms.isEmptyOrWhitespace || searchTerms.trimming().count < 2 {
            return nil
        } else {
            return NSPredicate.wordsWithinFields(searchTerms, fieldNames: #keyPath(Book.title), #keyPath(Book.authorSort), "ANY \(#keyPath(Book.subjects)).name")
        }
    }

    func updateSearchResults(for searchController: UISearchController) {
        let searchPredicate = getSearchPredicate()

        switch listBookSource! {
        case .controller(let controller):
            if let searchPredicate = searchPredicate {
                controller.fetchRequest.predicate = NSPredicate.and([defaultPredicate, searchPredicate])
            } else {
                controller.fetchRequest.predicate = defaultPredicate
            }
            try! controller.performFetch()
        case .orderedSet:
            if let searchPredicate = searchPredicate {
                listBookSource = .orderedSet(list.books.filtered(using: searchPredicate))
            } else {
                listBookSource = .orderedSet(list.books)
            }
        }
        tableView.reloadData()
    }
}

extension ListBookTable: DZNEmptyDataSetSource {
    func title(forEmptyDataSet scrollView: UIScrollView!) -> NSAttributedString! {
        if searchController.hasActiveSearchTerms {
            return title("🔍 No Results")
        }
        return title("✨ Empty List")
    }

    func verticalOffset(forEmptyDataSet scrollView: UIScrollView!) -> CGFloat {
        return -30
    }

    func description(forEmptyDataSet scrollView: UIScrollView!) -> NSAttributedString! {
        if searchController.hasActiveSearchTerms {
            return applyDescriptionAttributes(
                NSMutableAttributedString("Try changing your search, or add a new list by tapping the ", font: descriptionFont)
                .appending("+", font: boldDescriptionFont)
                .appending(" button above", font: descriptionFont)
            )
        }
        return applyDescriptionAttributes(
            NSMutableAttributedString("The list \"\(list.name)\" is currently empty.  To add a book to it, find a book and tap ", font: descriptionFont)
                .appending("Manage Lists", font: boldDescriptionFont)
        )
    }
}

extension ListBookTable: DZNEmptyDataSetDelegate {
    func emptyDataSetWillAppear(_ scrollView: UIScrollView!) {
        configureNavigationItem()

        // Prevents section headers lingering around for a bit after books are removed
        reloadHeaders()
    }

    func emptyDataSetDidDisappear(_ scrollView: UIScrollView!) {
        configureNavigationItem()
    }
}

extension ListBookTable: HeaderConfigurable {
    func configureHeader(_ header: UITableViewHeaderFooterView, at index: Int) {
        guard let header = header as? BookTableHeader else { preconditionFailure() }
        let numberOfRows = tableView.numberOfRows(inSection: index)
        if numberOfRows == 0 {
            header.removeFromSuperview()
        } else {
            header.configure(list: list, bookCount: numberOfRows, enableSort: !isEditing && !searchController.isActive)
        }
    }
}

extension ListBookTable: UITextFieldDelegate {
    func textFieldDidBeginEditing(_ textField: UITextField) {
        textField.text = list.name
    }

    func textFieldDidEndEditing(_ textField: UITextField) {
        textField.text = listNameFieldDefaultText
        // If we renamed the list, refresh the empty data set - if present
        if list.books.isEmpty {
            tableView.reloadData()
        }
    }

    func textFieldShouldReturn(_ textField: UITextField) -> Bool {
        guard let newText = textField.text, tryUpdateListName(to: newText) else { return false }
        textField.resignFirstResponder()
        return true
    }
}
